from .base_resources import ResourceList, ResourceCollection


class APIMetrics(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint='', filters=None):
        self.resources = (APIMetricPoints,)
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'metrics'


class APIMetricPoints(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'points'



from .api.client import Client
from .api.resources.admin_resources import AdminResources
from .api.resources.auth_resources import AuthResources
from .api.rehive_util import RehiveUtil
from .api.resources.public_resources import PublicResources
from .api.resources.user_resources import UserResources
from .api.resources.transaction_resource import (
    APITransactions, APITransactionCollections
)
from .api.resources.accounts_resources import APIAccounts
from .api.resources.company_resources import APICompany
from .api.resources.export_resources import APIExports
from .api.resources.metric_resources import APIMetrics
from .api.resources.account_definition_resources import APIAccountDefinitions
from .api.resources.document_type_resources import APIDocumentTypes
from .api.resources.group_resources import APIGroups


class Rehive:

    def __init__(self, token=None, connection_pool_size=0, network="live", debug=False, **kwargs):
        # API Classes
        self.debug = debug
        self.client = Client(
            token,
            connection_pool_size,
            network,
            debug,
            **kwargs
        )
        self.admin = AdminResources(self.client)
        self.auth = AuthResources(self.client)
        self.util = RehiveUtil(self.client)
        self.user = UserResources(self.client)
        self.transactions = APITransactions(self.client)
        self.transaction_collections = APITransactionCollections(self.client)
        self.accounts = APIAccounts(self.client)
        self.account_definitions = APIAccountDefinitions(self.client)
        self.company = APICompany(self.client)
        self.exports = APIExports(self.client)
        self.metrics = APIMetrics(self.client)
        self.public = PublicResources(self.client)
        self.document_types = APIDocumentTypes(self.client)
        self.groups = APIGroups(self.client)



from .base_resources import ResourceList


class APITransactions(ResourceList):

    def get_totals(self, **kwargs):
        response = self.get('totals/', **kwargs)
        return response

    def create_credit(self, amount, currency, **kwargs):
        data = {
            'amount': amount,
            'currency': currency
        }
        response = self.post(data, 'credit/', **kwargs)
        return response

    def create_debit(self, amount, currency, **kwargs):
        data = {
            'amount': amount,
            'currency': currency
        }
        response = self.post(data, 'debit/', **kwargs)
        return response

    def create_transfer(self, amount, recipient, currency, **kwargs):
        data = {
            'amount': amount,
            'recipient': recipient,
            'currency': currency
        }
        response = self.post(data, 'transfer/', **kwargs)
        return response

    @classmethod
    def get_resource_name(cls):
        return 'transactions'


class APITransactionCollections(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'transaction-collections'



from .base_resources import ResourceList, ResourceCollection


class APIPermissions(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'permissions'


class APIFees(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'fees'


class APIRequirements(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'requirements'


class APILimits(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'limits'


class APIRequirementItems(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'requirement-items'


class APIRequirementSets(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIRequirementItems
        )
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'requirement-sets'


class APITiers(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIFees,
            APIRequirements,
            APILimits,
            APIRequirementSets
        )
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'tiers'


class APIGroups(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIPermissions,
            APITiers,
            APIFees,
        )
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'groups'



from rehive.api.resources.base_resources import ResourceList, ResourceCollection


class PublicResources(ResourceList, ResourceCollection):
    def __init__(self, client):
        self.client = client
        self.endpoint = ''
        self.resources = (
            APILegalTerms,
            APICompanies
        )
        super().__init__(client, self.endpoint)
        self.create_resources(self.resources)

    @classmethod
    def get_resource_name(cls):
        return 'public'


class APIVersions(ResourceList):
    @classmethod
    def get_resource_name(cls):
        return 'versions'


class APILegalTerms(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint='', filters=None):
        self.resources = (APIVersions,)
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'legal-terms'


class APIGroups(ResourceList):
    @classmethod
    def get_resource_name(cls):
        return 'groups'


class APICompanies(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint='', filters=None):
        self.resources = (APIGroups, APILegalTerms)
        super().__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'companies'




import re
from io import BufferedReader

import copy
from urllib.parse import urlparse
from urllib.parse import urljoin
import urllib.parse
from rehive.api.exception import NoNextException, NoPreviousException


class Resource(object):
    def __init__(self, client, endpoint, filters=None):
        self.client = client
        self.endpoint = endpoint + self.get_resource_name() + '/'
        self.filters = filters
        self.resource_identifier = ''
        self.has_been_hydrated = False

    def get(self, resource_id=None, timeout=None, **kwargs):
        url = self._build_url(resource_id, **kwargs)
        response = self.client.get(url, timeout=timeout)
        return self._handle_resource_data(response)

    def post(self, data=None, resource_id=None, idempotent_key=None, timeout=None, **kwargs):
        # Allow us to parse through arbitrary request arguments
        if data is None:
            data = {}

        data, files = self._handle_file_data(**data, **kwargs)
        # We need this flag to force non-json on file uploads
        json = not bool(files) if data.get('json') is None else data.pop('json')
        url = self._build_url(resource_id)
        response = self.client.post(
            url,
            data,
            json=json,
            idempotent_key=idempotent_key,
            timeout=timeout,
            files=files
        )
        return self._handle_resource_data(response)

    def put(self, resource_id='', idempotent_key=None, timeout=None, **kwargs):
        url = self._build_url(resource_id)
        data, files = self._handle_file_data(**kwargs)
        # We need this flag to force non-json on file uploads
        json = not bool(files) if data.get('json') is None else data.pop('json')
        response = self.client.put(
            url,
            data,
            json=json,
            idempotent_key=idempotent_key,
            timeout=timeout,
            files=files
        )
        return self._handle_resource_data(response)

    def patch(self, resource_id='', idempotent_key=None, timeout=None, **kwargs):
        url = self._build_url(resource_id)
        data, files = self._handle_file_data(**kwargs)
        # We need this flag to force non-json on file uploads
        json = not bool(files) if data.get('json') is None else data.pop('json')
        response = self.client.patch(
            url,
            data,
            json=json,
            idempotent_key=idempotent_key,
            timeout=timeout,
            files=files
        )
        return self._handle_resource_data(response)

    def delete(self, resource_id='', timeout=None, **kwargs):
        data = kwargs
        url = self._build_url(resource_id)
        response = self.client.delete(url, data, timeout=timeout)
        return self._handle_resource_data(response)

    def options(self, resource_id='', timeout=None, **kwargs):
        url = self._build_url(resource_id)
        response = self.client.options(url, timeout=timeout)
        return self._handle_resource_data(response)

    def update(self, resource_id='', idempotent_key=None, timeout=None, **kwargs):
        return self.patch(
            resource_id,
            idempotent_key=idempotent_key,
            timeout=timeout,
            **kwargs
        )

    def create(self, idempotent_key=None, timeout=None, **kwargs):
        return self.post(
            idempotent_key=idempotent_key,
            timeout=timeout,
            **kwargs
        )

    # PRIVATE METHODS
    def _build_url(self, resource_id=None, **kwargs):
        # currently pagination should override all
        if kwargs.get('pagination'):
            endpoint = urllib.parse.urljoin(self.endpoint, kwargs.get('pagination'))
        else:
            if resource_id:
                endpoint = urllib.parse.urljoin(
                    self.endpoint, resource_id
                )
            else:
                endpoint = self.endpoint
            endpoint = self._append_trailing_slash(endpoint)
            if (self.resource_identifier is not None):
                endpoint = urllib.parse.urljoin(
                    endpoint, self.resource_identifier
                )
            endpoint = self._append_trailing_slash(endpoint)
        if kwargs.get('filters'):
            filters = kwargs.get('filters')
            filters_clean = {}

            # Filter out any special case filters
            array_type_filters = {}
            for f in filters:
                if type(filters[f]) is list:
                    array_type_filters[f] = filters[f]
                else:
                    filters_clean[f] = filters[f]

            array_filters_string = ''
            for array_filter in array_type_filters:
                encoded_array = urllib.parse.urlencode(
                    {array_filter: array_type_filters[array_filter]},
                    safe="[],'\""
                )
                array_filters_string = re.sub('[\[\'\]\+]', '', encoded_array)

            filters_string = urllib.parse.urlencode(filters_clean)
            if filters_string:
                filters_string = filters_string + '&' + array_filters_string
            else:
                filters_string = array_filters_string

            if "?" not in endpoint:
                endpoint = endpoint + '?' + filters_string
            else:
                endpoint = endpoint + '&' + filters_string
        return endpoint

    def _put_or_patch_by_identifier(self,
                                    method,
                                    identifier,
                                    identifier_field,
                                    **kwargs):
        if (identifier is not None and identifier_field is not None):
            resource_id = '?' + identifier_field + '=' + identifier
        else:
            raise Exception('No identifier supplied')
        data = kwargs
        url = self._build_url(resource_id)
        if method == 'PATCH':
            response = self.client.patch(url, data)
        elif method == 'PUT':
            response = self.client.put(url, data)
        else:
            raise Exception('No method supplied')
        return response

    def _append_trailing_slash(self, url):
        if (re.search(r'\/+$', url) is None):
            url = url + '/'
        return url

    def _set_resource_identifier(self, resource_identifiter):
        self.resource_identifier = str(resource_identifiter)

    def _set_endpoint(self, endpoint):
        self.endpoint = self._append_trailing_slash(endpoint)

    def get_resource_name(cls):
        raise NotImplementedError(
            'The resource should define its own string name'
        )

    def _handle_resource_data(self, response):
        return_data = response
        if 'data' in response:
            return_data = response.get('data')
        self.has_been_hydrated = True

        return return_data

    def _handle_file_data(self, **kwargs):
        data = {}
        files = {}
        for key, val in kwargs.items():
            if isinstance(val, BufferedReader):
                files[key] = val
            else:
                data[key] = val

        return data, files


class ResourceList(Resource):

    def __init__(self, client, endpoint, filters=None):
        super().__init__(client, endpoint, filters)
        self.next = None
        self.previous = None
        self._count = 0

    def get(self, endpoint=None, **kwargs):
        if endpoint:
            endpoint = str(endpoint)
        url = self._build_url(endpoint, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    def get_next(self, **kwargs):
        if self.next is None:
            raise NoNextException
        url = self._build_url(pagination=self.next, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    def get_previous(self, **kwargs):
        if self.previous is None:
            raise NoPreviousException
        url = self._build_url(pagination=self.previous, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    @property
    def count(self):
        if (self.has_been_hydrated is False):
            self.get()
        return self._count

    # PRIVATE METHODS
    def _handle_pagination_data(self, response):
        data = response.get('data')
        return_data = data
        if 'next' in data:
            if data.get('next') is not None:
                self.next = self._get_next_page_filter(data.get('next'))
            else:
                self.next = None
        if 'previous' in data:
            if data.get('previous') is not None:
                self.previous = self._get_next_page_filter(data['previous'])
            else:
                self.previous = None
        if 'count' in data:
            self._count = data.get('count')
        if 'results' in data:
            return_data = data.get('results')
        self.has_been_hydrated = True
        return return_data

    def _get_next_page_filter(self, string):
        url_segments = string.split('/')
        last_segment = url_segments[-1]
        return last_segment

    def _build_pagination_url(self, pagination, resource_id=None):
        endpoint = self.endpoint + resource_id if resource_id else self.endpoint
        paginatated_endpoint = endpoint + pagination
        if self.filters:
            paginatated_endpoint = paginatated_endpoint + self.filters
        return paginatated_endpoint


class ResourceCollection(object):

    def __init__(self):
        self.create_resources(self.resources)

    def create_resources(self, resources):
        for resource in resources:
            setattr(
                self,
                self._sanitize_resource_name(
                    resource.get_resource_name()
                ),
                resource(self.client, self.endpoint)
            )

    def obj(self, resource_identifiter):
        return self.object(resource_identifiter)

    def object(self, resource_identifiter):
        resource_object = copy.copy(self)
        resource_object._set_resource_identifier(resource_identifiter)
        resource_object._set_endpoint(
            resource_object.endpoint + resource_object.resource_identifier
        )
        resource_object.create_resources(resource_object.resources)
        return resource_object

    def _sanitize_resource_name(self, string):
        return string.replace('-', '_')
